# Best possible skill rating
ELITE_PLAYER_SKILL_LEVEL = 1

DAY_MAP = {
    "Monday": "M",
    "Tuesday": "T",
    "Wednesday": "W",
    "Thursday": "R",
    "Friday": "F",
}

FIELD_MAP = {
    "East": ["Ahern", "Donnelly"],
    "Central": ["Common"],
    "Cambridgeport": ["Pacific", "Magazine"],
    "West": ["Danehy", "Raymond", "Maher"],
    "North": ["Danehy", "Raymond", "Russell"],
}
